unit ZlawMySQLBackup;

interface

uses
  SysUtils, Classes, ZConnection, ZlawConnection, ZDataset;

type
  TZlawMySQLBackupOption=(lmbDropCreateDatabase,lmbUseDatabase,lmbFlushLogsBeforeDump,
    lmbSetKeyChecksOff,lmbDropObjects,lmbLockAroundInserts,lmbDumpData);
  TZlawMySQLBackupOptions=set of TZlawMySQLBackupOption;

  TZlawMySQLBackupObject=(lmboTables,lmboViews,lmboProcedures,lmboFunctions,lmboTriggers);
  TZlawMySQLBackupObjects=set of TZlawMySQLBackupObject;

  TZlawMySQLBackupNotifyEvent=procedure(const ObjectName:string) of object;
  TZlawMySQLBackupDumpDataEvent=procedure(const RecNo,RecCount:integer) of object;

  TZlawMySQLBackup = class(TZlawConnection)
  private
    { Private declarations }
    FQuery,FTables:TZReadOnlyQuery;
    FFileStream:TFileStream;
    FMemStream:TMemoryStream;
    FTableName:string;
    FBuffer:string;
    FObjects: TZlawMySQLBackupObjects;
    FOptions: TZlawMySQLBackupOptions;
    FOnWriteFooter: TNotifyEvent;
    FOnWriteHeader: TNotifyEvent;
    FOnWriteTableData: TZlawMySQLBackupDumpDataEvent;
    FOnWriteTableDef: TZlawMySQLBackupNotifyEvent;
    FOnWriteTriggerDef: TZlawMySQLBackupNotifyEvent;
    FOnWriteProcedureDef: TZlawMySQLBackupNotifyEvent;
    FOnWriteViewDef: TZlawMySQLBackupNotifyEvent;
    procedure SetObjects(const Value: TZlawMySQLBackupObjects);
    procedure SetOptions(const Value: TZlawMySQLBackupOptions);
    procedure SetOnWriteFooter(const Value: TNotifyEvent);
    procedure SetOnWriteHeader(const Value: TNotifyEvent);
    procedure SetOnWriteProcedureDef(
      const Value: TZlawMySQLBackupNotifyEvent);
    procedure SetOnWriteTableData(
      const Value: TZlawMySQLBackupDumpDataEvent);
    procedure SetOnWriteTableDef(const Value: TZlawMySQLBackupNotifyEvent);
    procedure SetOnWriteTriggerDef(
      const Value: TZlawMySQLBackupNotifyEvent);
    procedure SetOnWriteViewDef(const Value: TZlawMySQLBackupNotifyEvent);
  protected
    { Protected declarations }
    procedure WriteHeader;
    procedure WriteFooter;
    procedure WriteTableDefs;
    procedure WriteTableData(const tablename:string);
    procedure WriteViewDefs;
    procedure WriteProcedureDefs;
    procedure WriteTriggerDefs;
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Backup(const filename:string);
    property TableName:string read FTableName write FTableName;
  published
    { Published declarations }
    property Options:TZlawMySQLBackupOptions read FOptions write SetOptions;
    property Objects:TZlawMySQLBackupObjects read FObjects write SetObjects;
    property OnWriteHeader:TNotifyEvent read FOnWriteHeader write SetOnWriteHeader;
    property OnWriteTableDef:TZlawMySQLBackupNotifyEvent read FOnWriteTableDef write SetOnWriteTableDef;
    property OnWriteTableData:TZlawMySQLBackupDumpDataEvent read FOnWriteTableData write SetOnWriteTableData;
    property OnWriteViewDef:TZlawMySQLBackupNotifyEvent read FOnWriteViewDef write SetOnWriteViewDef;
    property OnWriteProcedureDef:TZlawMySQLBackupNotifyEvent read FOnWriteProcedureDef write SetOnWriteProcedureDef;
    property OnWriteTriggerDef:TZlawMySQLBackupNotifyEvent read FOnWriteTriggerDef write SetOnWriteTriggerDef;
    property OnWriteFooter:TNotifyEvent read FOnWriteFooter write SetOnWriteFooter;
  end;

procedure Register;

implementation

uses DB, DateUtils, lawStr, lawStream;

const crlf=#13#10;

procedure Register;
begin
  RegisterComponents('Zeos Access', [TZlawMySQLBackup]);
end;

{ TZlawMySQLBackup }
constructor TZlawMySQLBackup.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Objects:=[lmboTables,lmboViews,lmboProcedures,lmboFunctions,lmboTriggers];
  Options:=[lmbDropCreateDatabase,lmbUseDatabase,lmbFlushLogsBeforeDump,
    lmbSetKeyChecksOff,lmbDropObjects,lmbLockAroundInserts,lmbDumpData];
  FTables:=TZReadOnlyQuery.Create(self);
  FTables.Connection:=self;
  FQuery:=TZReadOnlyQuery.Create(self);
  FQuery.Connection:=self;
end;

destructor TZlawMySQLBackup.Destroy;
begin
  FTables.Free;
  FQuery.Free;
  inherited Destroy;
end;

procedure TZlawMySQLBackup.WriteHeader;
begin
  FBuffer:='';
  if lmbSetKeyChecksOff in FOptions
  then FBuffer:=FBuffer+'SET @OLD_UNIQUE_CHECKS=@@UNIQUE_CHECKS, UNIQUE_CHECKS=0'+crlf+crlf+
    'SET @OLD_FOREIGN_KEY_CHECKS=@@FOREIGN_KEY_CHECKS, FOREIGN_KEY_CHECKS=0'+crlf+crlf;
  FBuffer:=FBuffer+'SET @OLD_SQL_MODE=@@SQL_MODE, SQL_MODE=''NO_AUTO_VALUE_ON_ZERO'''+crlf+crlf;
  if lmbDropCreateDatabase in FOptions
  then FBuffer:=FBuffer+'DROP DATABASE IF EXISTS '+Database+crlf+crlf+
    'CREATE DATABASE '+Database+crlf+crlf;
  if lmbUseDatabase in FOptions
  then FBuffer:=FBuffer+'USE '+Database+crlf+crlf;
  FFileStream.Write(FBuffer[1],Length(FBuffer));
  if Assigned(FOnWriteHeader)
  then FOnWriteHeader(self);
end;

procedure TZlawMySQLBackup.WriteFooter;
begin
  FBuffer:='SET SQL_MODE=@OLD_SQL_MODE'+crlf+crlf;
  if lmbSetKeyChecksOff in FOptions
  then FBuffer:=FBuffer+'SET FOREIGN_KEY_CHECKS=@OLD_FOREIGN_KEY_CHECKS'+crlf+crlf+
    'SET UNIQUE_CHECKS=@OLD_UNIQUE_CHECKS'+crlf+crlf;
  FFileStream.Write(FBuffer[1],Length(FBuffer));
  if Assigned(FOnWriteFooter)
  then FOnWriteFooter(self);
end;

procedure TZlawMySQLBackup.WriteTableDefs;
begin
  if lmboTables in FObjects
  then begin
    FTables.SQL.Text:='SHOW FULL TABLES WHERE Table_type=''BASE TABLE''';
    FTables.Open;
    while not FTables.Eof do begin
      if (FTables.Fields[0].AsString=TableName) or (TableName='')
      then begin
        FBuffer:='';
        if lmbDropObjects in FOptions
        then FBuffer:=FBuffer+'DROP TABLE IF EXISTS '+FTables.Fields[0].AsString+crlf+crlf;
        FQuery.SQL.Text:='SHOW CREATE TABLE '+FTables.Fields[0].AsString;
        FQuery.Open;
        FBuffer:=FBuffer+FQuery.Fields[1].AsString+crlf+crlf;
        FFileStream.Write(FBuffer[1],Length(FBuffer));
        if Assigned(FOnWriteTableDef)
        then FOnWriteTableDef(FTables.Fields[0].AsString);
        WriteTableData(FTables.Fields[0].AsString);
      end;
      FTables.Next;
    end;
  end;
end;

procedure TZlawMySQLBackup.WriteTableData(const tablename:string);
var comma:string;
    cikl,reccount:integer;
begin
  if lmbDumpData in FOptions
  then begin
    FBuffer:='';
    if lmbLockAroundInserts in FOptions
    then FBuffer:=FBuffer+'LOCK TABLES '+tablename+' WRITE'+crlf+crlf;
    FBuffer:=FBuffer+'ALTER TABLE '+tablename+' DISABLE KEYS'+crlf+crlf;
    FFileStream.Write(FBuffer[1],Length(FBuffer));
    FQuery.SQL.Text:='SELECT COUNT(*) FROM '+tablename;
    FQuery.Open;
    reccount:=FQuery.Fields[0].AsInteger;
    if reccount>0
    then begin
      FBuffer:='INSERT INTO '+tablename+' VALUES ';
      FFileStream.Write(FBuffer[1],Length(FBuffer));
      FQuery.SQL.Text:='SELECT * FROM '+tablename;
      FQuery.Open;
      while not FQuery.Eof do begin
        FBuffer:='(';
        FFileStream.Write(FBuffer[1],Length(FBuffer));
        for cikl:=0 to FQuery.FieldCount-1 do begin
          if cikl=FQuery.FieldCount-1
          then comma:=''
          else comma:=',';
          if FQuery.Fields[cikl].IsNull
          then begin
            FBuffer:='NULL'+comma;
            FFileStream.Write(FBuffer[1],Length(FBuffer));
          end
          else
            case FQuery.Fields[cikl].DataType of
              ftString:begin
                FBuffer:=''''+UnEscapeString(FQuery.Fields[cikl].AsString,'\')+''''+comma;
                FFileStream.Write(FBuffer[1],Length(FBuffer));
              end;
              ftSmallInt,ftInteger,ftWord,ftLargeInt:begin
                FBuffer:=IntToStrZero(FQuery.Fields[cikl].AsInteger,1)+comma;
                FFileStream.Write(FBuffer[1],Length(FBuffer));
              end;
              ftFloat,ftCurrency:begin
                FBuffer:=RS(FormatFloat('0',FQuery.Fields[cikl].AsFloat),',','.')+comma;
                FFileStream.Write(FBuffer[1],Length(FBuffer));
              end;
              ftDate:begin
                FBuffer:=''''+IntToStr(YearOf(FQuery.Fields[cikl].AsDateTime))+'-'+IntToStrZero(MonthOf(FQuery.Fields[cikl].AsDateTime),2)+'-'+
                  IntToStrZero(DayOf(FQuery.Fields[cikl].AsDateTime),2)+''''+comma;
                FFileStream.Write(FBuffer[1],Length(FBuffer));
              end;
              ftDateTime:begin
                FBuffer:=''''+IntToStr(YearOf(FQuery.Fields[cikl].AsDateTime))+'-'+IntToStrZero(MonthOf(FQuery.Fields[cikl].AsDateTime),2)+'-'+
                  IntToStrZero(DayOf(FQuery.Fields[cikl].AsDateTime),2)+' '+IntToStrZero(HourOf(FQuery.Fields[cikl].AsDateTime),2)+':'+
                  IntToStrZero(MinuteOf(FQuery.Fields[cikl].AsDateTime),2)+':'+IntToStrZero(SecondOf(FQuery.Fields[cikl].AsDateTime),2)+''''+comma;
                FFileStream.Write(FBuffer[1],Length(FBuffer));
              end;
              ftBlob:begin
                FBuffer:='0x';
                FFileStream.Write(FBuffer[1],Length(FBuffer));
                FMemStream.Clear;
                TBlobField(FQuery.Fields[cikl]).SaveToStream(FMemStream);
                ConvertStreamToHex(FMemStream,FFileStream);
                FBuffer:=comma;
                FFileStream.Write(FBuffer[1],Length(FBuffer));
              end;
              ftMemo:begin
                FBuffer:=''''+UnEscapeString(FQuery.Fields[cikl].AsString,'\')+''''+comma;
                FFileStream.Write(FBuffer[1],Length(FBuffer));
              end;
              else begin
                FBuffer:=IntToStrZero(Ord(FQuery.Fields[cikl].DataType),2)+comma;
                FFileStream.Write(FBuffer[1],Length(FBuffer));
              end;
            end;
        end;
        if Assigned(FOnWriteTableData)
        then FOnWriteTableData(FQuery.RecNo,reccount);
        FQuery.Next;
        if FQuery.Eof
        then FBuffer:=')'+crlf+crlf
        else FBuffer:='),';
        FFileStream.Write(FBuffer[1],Length(FBuffer));
      end;
    end;
    FBuffer:='ALTER TABLE '+tablename+' ENABLE KEYS'+crlf+crlf;
    if lmbLockAroundInserts in FOptions
    then FBuffer:=FBuffer+'UNLOCK TABLES'+crlf+crlf;
    FFileStream.Write(FBuffer[1],Length(FBuffer));
  end;
end;

procedure TZlawMySQLBackup.WriteViewDefs;
begin
  if lmboViews in FObjects
  then begin
    FTables.SQL.Text:='SHOW FULL TABLES WHERE Table_type=''VIEW''';
    FTables.Open;
    while not FTables.Eof do begin
      FBuffer:='';
      if lmbDropObjects in FOptions
      then FBuffer:=FBuffer+'DROP VIEW IF EXISTS '+FTables.Fields[0].AsString+crlf+crlf;
      FQuery.SQL.Text:='SHOW CREATE VIEW '+FTables.Fields[0].AsString;
      FQuery.Open;
      FBuffer:=FBuffer+FQuery.Fields[1].AsString+crlf+crlf;
      FFileStream.Write(FBuffer[1],Length(FBuffer));
      if Assigned(FOnWriteViewDef)
      then FOnWriteViewDef(FTables.Fields[0].AsString);
      FTables.Next;
    end;
  end;
end;

procedure TZlawMySQLBackup.WriteProcedureDefs;
var db:string;
    proclist,funclist:TStringList;
    cikl:integer;
begin
  if (lmboProcedures in FObjects) or (lmboFunctions in FObjects)
  then begin
    proclist:=TStringList.Create;
    funclist:=TStringList.Create;
    try
      db:=Database;
      Disconnect;
      Database:='information_schema';
      Connect;
      FTables.SQL.Text:='SELECT * FROM ROUTINES WHERE ROUTINE_SCHEMA='''+db+'''';
      FTables.Open;
      while not FTables.Eof do begin
        if FTables.FieldByName('ROUTINE_TYPE').AsString='PROCEDURE'
        then proclist.Add(FTables.FieldByName('ROUTINE_NAME').AsString)
        else
        if FTables.FieldByName('ROUTINE_TYPE').AsString='FUNCTION'
        then funclist.Add(FTables.FieldByName('ROUTINE_NAME').AsString);
        FTables.Next;
      end;
      Disconnect;
      Database:=db;
      Connect;
      if lmboProcedures in FObjects
      then begin
        for cikl:=0 to proclist.Count-1 do begin
          FTables.SQL.Text:='SHOW CREATE PROCEDURE '+proclist[cikl];
          FTables.Open;
          FBuffer:='';
          if lmbDropObjects in FOptions
          then FBuffer:=FBuffer+'DROP PROCEDURE IF EXISTS '+proclist[cikl]+crlf+crlf;
          FBuffer:=FBuffer+FTables.Fields[2].AsString+crlf+crlf;
          FFileStream.Write(FBuffer[1],Length(FBuffer));
          if Assigned(FOnWriteProcedureDef)
          then FOnWriteProcedureDef(proclist[cikl]);
        end;
      end;
      if lmboFunctions in FObjects
      then begin
        for cikl:=0 to funclist.Count-1 do begin
          FTables.SQL.Text:='SHOW CREATE FUNCTION '+funclist[cikl];
          FTables.Open;
          FBuffer:='';
          if lmbDropObjects in FOptions
          then FBuffer:=Fbuffer+'DROP FUNCTION IF EXISTS '+funclist[cikl]+crlf+crlf;
          FBuffer:=FBuffer+FTables.Fields[2].AsString+crlf+crlf;
          FFileStream.Write(FBuffer[1],Length(FBuffer));
          if Assigned(FOnWriteProcedureDef)
          then FOnWriteProcedureDef(proclist[cikl]);
        end;
      end;
    finally
      funclist.Free;
      proclist.Free;
    end;
  end;
end;

procedure TZlawMySQLBackup.WriteTriggerDefs;
begin
  if lmboTriggers in FObjects
  then begin
    FTables.SQL.Text:='SHOW TRIGGERS';
    FTables.Open;
    while not FTables.Eof do begin
      FBuffer:='';
      if lmbDropObjects in FOptions
      then FBuffer:=FBuffer+'DROP TRIGGER IF EXISTS '+FTables.FieldByName('Trigger').AsString+crlf+crlf;
      FBuffer:=FBuffer+'CREATE TRIGGER '+FTables.FieldByName('Trigger').AsString+' '+
        FTables.FieldByName('Timing').AsString+' '+FTables.FieldByName('Event').AsString+' ON '+
        FTables.FieldByName('Table').AsString+' '+
        'FOR EACH ROW '+FTables.FieldByName('Statement').AsString+crlf+crlf;
      FFileStream.Write(FBuffer[1],Length(FBuffer));
      if Assigned(FOnWriteTriggerDef)
      then FOnWriteTriggerDef(FTables.FieldByName('Trigger').AsString);
      FTables.Next;
    end;
  end;
end;

procedure TZlawMySQLBackup.Backup(const filename:string);
begin
  FFileStream:=TFileStream.Create(filename,fmCreate);
  FMemStream:=TMemoryStream.Create;
  try
    if lmbFlushLogsBeforeDump in Options
    then begin
      FQuery.SQL.Text:='FLUSH LOGS';
      FQuery.ExecSQL;
    end;
    WriteHeader;
    WriteTableDefs;
    WriteViewDefs;
    WriteProcedureDefs;
    WriteTriggerDefs;
    WriteFooter;
  finally
    FMemStream.Free;
    FFileStream.Free;
  end;
end;

procedure TZlawMySQLBackup.SetObjects(
  const Value: TZlawMySQLBackupObjects);
begin
  FObjects := Value;
end;

procedure TZlawMySQLBackup.SetOptions(
  const Value: TZlawMySQLBackupOptions);
begin
  FOptions := Value;
end;

procedure TZlawMySQLBackup.SetOnWriteFooter(const Value: TNotifyEvent);
begin
  FOnWriteFooter := Value;
end;

procedure TZlawMySQLBackup.SetOnWriteHeader(const Value: TNotifyEvent);
begin
  FOnWriteHeader := Value;
end;

procedure TZlawMySQLBackup.SetOnWriteProcedureDef(
  const Value: TZlawMySQLBackupNotifyEvent);
begin
  FOnWriteProcedureDef := Value;
end;

procedure TZlawMySQLBackup.SetOnWriteTableData(
  const Value: TZlawMySQLBackupDumpDataEvent);
begin
  FOnWriteTableData := Value;
end;

procedure TZlawMySQLBackup.SetOnWriteTableDef(
  const Value: TZlawMySQLBackupNotifyEvent);
begin
  FOnWriteTableDef := Value;
end;

procedure TZlawMySQLBackup.SetOnWriteTriggerDef(
  const Value: TZlawMySQLBackupNotifyEvent);
begin
  FOnWriteTriggerDef := Value;
end;

procedure TZlawMySQLBackup.SetOnWriteViewDef(
  const Value: TZlawMySQLBackupNotifyEvent);
begin
  FOnWriteViewDef := Value;
end;

end.
